;- Window Constants
#WindowMain = 0
#WindowMainWidth = 484
#WindowMainHeight = 445

;- MenuBar Constants
#MenuBar     = 0
#mNEW        = 1
#mEASY       = 2
#mMEDIUM     = 3
#mHARD       = 4
#mSOUND      = 5
#mQUIT       = 6
#mABOUT      = 7

;- StatusBar Constants
#StatusBar = 0

;- Game constants
#RED            = 0
#YELLOW         = 1
#BOARD          = 2
#SOUND          = 3
#ICON           = 4
#WINPLACES      = 69
#MAXPIECES      = 42
#EMPTY          = 2
#MAXSTATE       = 10

;- Globals
Global SoundOn.b, GameOver.b, MyMove.b, GLevel.b, Pos.b
Global ComputerPoint.Point

Dim Map.b(7,6,#WINPLACES)
Dim Score.w(#MAXSTATE, 2, #WINPLACES)
Dim Board.b(#MAXSTATE, 6, 7)
Dim NumPieces.w(#MAXSTATE)

Procedure OpenWindowMain()
  Shared SoundOn
  UseJPEGImageDecoder()
  CatchImage(#RED, ?RedCoin)
  CatchImage(#YELLOW, ?YellowCoin)
  CatchImage(#BOARD, ?Connect4Board)
  CatchImage(#ICON, ?Connect4Icon)

  If OpenWindow(#WindowMain, 297, 167, #WindowMainWidth, #WindowMainHeight, #PB_Window_SystemMenu | #PB_Window_TitleBar | #PB_Window_ScreenCentered , "Pure Connect4")
    If CreateMenu(#MenuBar, WindowID())
      MenuTitle("&Game")
      MenuItem(#mNEW, "&New          F2")
      MenuBar()
      MenuItem(#mEASY, "&Easy")
      MenuItem(#mMEDIUM, "&Medium")
      MenuItem(#mHARD, "&Hard")
      MenuBar()
      MenuItem(#mSOUND, "&Sound")
      MenuBar()
      MenuItem(#mQUIT, "&Quit")
      MenuTitle("&Help")
      MenuItem(#mABOUT, "&About")     
    EndIf
    
    If CreateStatusBar(#StatusBar, WindowID())
      AddStatusBarField(#WindowMainWidth)
      StatusBarText(#StatusBar, 0, "Ready.")
    EndIf

    If InitSound() = 0
      ; Disable sound menu 
      DisableMenuItem(#mSOUND,1)
      SoundOn = #FALSE
    Else
      CatchSound(#SOUND, ?DropSound)
      SoundOn = #TRUE
      SetMenuItemState(#MenuBar, #mSOUND, 1)
    EndIf
    
    AddKeyboardShortcut(#WindowMain, #PB_Shortcut_F2, #mNEW) 
    
    glevel = 3
    gameOver = #TRUE
    SetMenuItemState(#MenuBar, #mEASY, 1)
        
  EndIf
EndProcedure

Procedure Paint()
  StartDrawing(WindowOutput())  
  
  DrawImage(UseImage(#BOARD), 20, 10)  
  For i = 0 To 6
    For j = 0 To 5
      Select board(0, i, j)
        Case 0
          DrawImage(UseImage(#RED), 28 + i * 62, 328 - j *62)
        Case 1
          DrawImage(UseImage(#YELLOW), 28 + i * 62, 328 - j *62)
      EndSelect                  
    Next
  Next
  
  StopDrawing()
EndProcedure


Procedure init()
   count.w
  
  For i = 0 To 6
    For j =0 To 5
      For k = 0 To #winPlaces - 1
        map(i, j, k) = #FALSE
      Next
    Next
  Next
  
  ; Set the horizontal win positions
  For i = 0 To 5
    For j = 0 To 3
      For k = 0 To 3
        map(j + k, i, count) = #TRUE
      Next
      count = count + 1
    Next
  Next
  
  ; Set the vertical win positions
  For i = 0 To 6
    For j = 0 To 2
      For k = 0 To 3
        map(i, j + k, count) = #TRUE
      Next
      count = count + 1
    Next
  Next  
  
  ; Set the forward diagonal win positions
  For i = 0 To 2
    For j = 0 To 3
      For k = 0 To 3
        map(j + k, i + k, count) = #TRUE
      Next
      count = count + 1
    Next
  Next
  
  ; Set the backward diagonal win positions
  For i = 0 To 2
    For j = 6 To 3 Step -1
      For k = 0 To 3
        map(j - k, i + k, count) = #TRUE
      Next
      count = count + 1
    Next
  Next

  ; Initialize the board
  For i = 0 To 6
    For j = 0 To 5
      board(0,i,j) = #empty
    Next
  Next
  
  ; Initialize the scores
  For i = 0 To 1
    For j = 0 To #winPlaces - 1
      score(0, i, j) = 1
    Next
  Next
  
  numPieces(0) = 0
EndProcedure

Procedure WindowCallback(WindowID, Message, lParam, wParam)
  If Message = #WM_PAINT
    Paint()
  EndIf
  
  ProcedureReturn #PB_ProcessPureBasicEvents
EndProcedure

Procedure newgame()
  GameOver = #FALSE
  MyMove = #TRUE
  StatusBarText(#StatusBar, 0, "Your turn.")
  init()
  paint()
EndProcedure

Procedure stateCopy(statefrom.w, stateto.w)
  For i = 0 To 6
    For j = 0 To 5
      board(stateto, i ,j) = board(statefrom, i ,j)
    Next
  Next
  
  For i = 0 To 1
    For j = 0 To #winPlaces - 1
      score(stateto, i, j)  = score(statefrom, i ,j)
    Next
  Next
   
  numPieces(stateto) = numPieces(statefrom)
EndProcedure

Procedure.b isWinner(state.w, player.b)  
  For i = 0 To #winPlaces - 1
    If score(state, player, i) = 16
      ProcedureReturn #TRUE
    EndIf
  Next
  
  ProcedureReturn #FALSE
EndProcedure

Procedure.b isTie()
  If numPieces(0) = #maxPieces
    ProcedureReturn #TRUE
  Else
    ProcedureReturn #FALSE
  EndIf
EndProcedure

Procedure.b getOtherPlayer(player.b)
  ProcedureReturn (1 - player)
EndProcedure

Procedure.w calcScore(state.w, player.b)
  s.w
  s = 0
  
  For i = 0 To #winPlaces - 1
    s = s + score(state, player, i)
  Next
  
  ProcedureReturn s
EndProcedure

Procedure updateScore(state.w, player.b, x.b, y.b)
 
  For i = 0 To #winPlaces - 1
    If map(x,y,i) = #TRUE
      score(state, player, i) = score(state, player, i) << 1
      score(state, getOtherPlayer(player), i) = 0
    EndIf
  Next
EndProcedure
     
Procedure.b dropPiece(state.w, player.b, xPos.b) 
  yPos.b
  yPos = 0
 
  While board(state, xPos, yPos) <> #empty 
    yPos = yPos + 1
    ;The column is full
    If yPos = 6
      ProcedureReturn -1
    EndIf
  Wend
  
  ; The move is OK
  board(state, xPos, yPos) = player
  numPieces(state) = numPieces(state) + 1
  updateScore(state, player, xPos, yPos)

  ProcedureReturn yPos
EndProcedure

Procedure.w evaluate(state.w, player.b, level.b, depth.b, alpha.w, beta.w)
  goodness.w
  best.w
  maxab = alpha

  If (level <> depth)
    best = - 30000
    For i = 0 To 6
      stateCopy(state, state + 1)
      If dropPiece(state + 1, getOtherPlayer(player), i) < 0
      
      Else
        If isWinner(state + 1, getOtherPlayer(player)) = #TRUE
          goodness = 25000 - depth
        Else 
          goodness = evaluate(state +1, getOtherPlayer(player), level, depth + 1, -beta, -maxab)
        EndIf
        If goodness > best
          best = goodness
          If best > maxab
            maxab = best
          EndIf
        EndIf
        If best > beta
          i = 6
        EndIf
            
      EndIf
    Next
    ProcedureReturn -best
  EndIf
  
  ProcedureReturn (calcscore(state, player) - calcscore(state, getotherplayer(player)))
EndProcedure

Procedure.b MakeMove(player.b, xPos.b)
  yPos.b
  yPos = dropPiece(0, player, xPos) 
  ProcedureReturn yPos
EndProcedure

Procedure.l computerMove(player.b, level.b)
  bestXPos.b 
  bestXpos = -1
  goodness = 0
  bestWorst = -30000 
  numOfEqual = 0

  ; Simulate a drop in each of the columns
  For i = 0 To 6
    stateCopy(0, 1)
   
    ; If column is full, move on
    If (dropPiece(1,player, i) < 0)
      ; continue
    Else

      ; If this drop wins the game, then cool 
      If (isWinner(1, player)) = #TRUE
        bestWorst = 25000
        bestXPos = i
    
      ; Otherwise, look ahead To see how good it is
      Else
        goodness = evaluate(1, player, level, 1, -30000, -bestWorst)

        ; If this move looks better than previous moves, remember it
        If (goodness > bestWorst) 
          bestWorst = goodness;
          bestXPos = i;
          numOfEqual = 1;
        EndIf

        ; If two moves are equally good, make a random choice
        If (goodness = bestWorst) 
          numOfEqual = numOfEqual + 1  
          If Random(numOfEqual) = numOfEqual
            bestXPos = i
          EndIf
        EndIf
      EndIf 
    EndIf
  Next

  ; Drop the piece in the best column
  If (bestXPos >= 0) 
    yPos = dropPiece(state, player, bestXPos)
    If (yPos >= 0)
      ComputerPoint\y = yPos
      ComputerPoint\x = bestXPos
      ProcedureReturn bestXPos
    EndIf
  EndIf
  ProcedureReturn bestXPos
EndProcedure 

Procedure freeresources()
  FreeSound(#SOUND)
  FreeImage(#BOARD)
  FreeImage(#RED)
  FreeImage(#YELLOW)
  End
EndProcedure

Procedure About()
  info.s
  info = "Pure Connect 4" 
  info = info + " (Verion 1.0)" + Chr(10) + Chr(13)
  info = info + "Written in Pure Basic"
  info = info + " by Lucky Luke"
  
  ShellAbout_(WindowID(), "Pure Connect4"+" # "+"Operating System", info , UseImage(#ICON))
EndProcedure

OpenWindowMain()

NewGame()

SetWindowCallback(@WindowCallback())

Repeat
  
  Event = WaitWindowEvent()
  
  If Event = #PB_EventMenu
        
    MenuID = EventMenuID()
    
    Select MenuID
      Case #mSOUND
        If SoundOn = #TRUE
          SoundOn = #FALSE
          SetMenuItemState(#MenuBar, #mSOUND, 0)
        Else
          SoundOn = #TRUE
          SetMenuItemState(#MenuBar, #mSOUND, 1)        
        EndIf
      Case #mEASY
        glevel = 3
        SetMenuItemState(#MenuBar, #mHARD, 0)
        SetMenuItemState(#MenuBar, #mMEDIUM, 0)
        SetMenuItemState(#MenuBar, #mEASY, 1)
        NewGame()
     Case #mMEDIUM
        glevel = 5
        SetMenuItemState(#MenuBar, #mHARD, 0)
        SetMenuItemState(#MenuBar, #mMEDIUM, 1)
        SetMenuItemState(#MenuBar, #mEASY, 0)
        NewGame()
     Case #mHARD
        glevel = 7
        SetMenuItemState(#MenuBar, #mHARD, 1)
        SetMenuItemState(#MenuBar, #mMEDIUM, 0)
        SetMenuItemState(#MenuBar, #mEASY, 0)
        NewGame()
      Case #mNEW
        NewGame()  
      Case #mQUIT
        freeresources()
      Case #mABOUT
        about()  
    EndSelect
  EndIf  
  If GameOver <> #TRUE
    If myMove = #TRUE  
      If event = 513 
        xmouse = WindowMouseX()    
        If xmouse > 35 And xmouse < 87
          pos = makemove(0,0)
        ElseIf xmouse > 98 And xmouse < 148
          pos = makemove(0,1)  
        ElseIf xmouse > 159 And xmouse < 210
          pos = makemove(0,2)  
        ElseIf xmouse > 222 And xmouse < 272
          pos = makemove(0,3)  
        ElseIf xmouse > 284 And xmouse < 334
          pos = makemove(0,4)  
        ElseIf xmouse > 346 And xmouse < 397
          pos = makemove(0,5)       
        ElseIf xmouse > 408 And xmouse < 458
          pos = makemove(0,6)  
        Else
          pos = -1
        EndIf
                      
        If pos >= 0          
          If isWinner(0, 0) = #FALSE
            If isTie() = #FALSE
              myMove = #FALSE
              StatusBarText(#StatusBar, 0, "Thinking...") 
            Else
              StatusBarText(#StatusBar, 0, "It's a tie.")
              gameover = #TRUE
            EndIf
          Else
            StatusBarText(#StatusBar, 0, "You won!")
            gameover = #TRUE
          EndIf
          Paint()   
          
          If SoundOn = #TRUE
            PlaySound(#SOUND, 0)
            Delay(1000)
          EndIf
                 
        EndIf      
      EndIf
    EndIf
    If myMove = #FALSE
      pos = computerMove(1, glevel)

      If ComputerPoint\y >= 0
        If isWinner(0, 1) = #FALSE
          If isTie() = #FALSE
            myMove = #TRUE
            StatusBarText(#StatusBar, 0, "Your turn.")
          Else
            StatusBarText(#StatusBar, 0, "It's a tie.")
            gameover = #TRUE
            myMove=#TRUE
          EndIf
        Else
          StatusBarText(#StatusBar, 0, "You lost!")
          gameover = #TRUE
          myMove = #TRUE
        EndIf
        paint() 
        
        If SoundOn = #TRUE
          PlaySound(#SOUND, 0)
        EndIf    
     EndIf  
    EndIf
  EndIf
  
Until event = #PB_Event_CloseWindow

End

; resources
DropSound: IncludeBinary "drop.wav" 
RedCoin: IncludeBinary "red.jpg"
YellowCoin: IncludeBinary "yellow.jpg"
Connect4Board: IncludeBinary "board.jpg"
Connect4Icon: IncludeBinary "connect4.ico"


; ExecutableFormat=
; UseIcon=C:\Documents and Settings\Luk\Bureaublad\connect4.ico
; Executable=C:\Documents and Settings\Luk\Bureaublad\Pure Connect.exe
; EOF